/*
 * $Id: FrequencySeries.m4,v 1.7 2008/09/17 18:00:00 kipp Exp $
 *
 * Copyright (C) 2007  Kipp Cannon
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */


#include <math.h>
#include <string.h>
#include <lal/Date.h>
#include <lal/LALComplex.h>
#include <lal/LALDatatypes.h>
#include <lal/LALStdlib.h>
#include <lal/LALStatusMacros.h>
#include <lal/LALErrno.h>
#include <lal/LALError.h>
#include <lal/FrequencySeries.h>
#include <lal/Sequence.h>
#include <lal/Units.h>
#include <lal/XLALError.h>


#include <lal/LALRCSID.h>
NRCSID(FREQUENCYSERIESC, "$Id: FrequencySeries.m4,v 1.7 2008/09/17 18:00:00 kipp Exp $");






void XLALDestroyREAL4FrequencySeries (
	REAL4FrequencySeries *series
)
{
	if(series)
		XLALDestroyREAL4Sequence (series->data);
	XLALFree(series);
}


void LALDestroyREAL4FrequencySeries (
	LALStatus *status,
	REAL4FrequencySeries *series
)
{
	INITSTATUS(status, "LALDestroyREAL4FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALDestroyREAL4FrequencySeries", "XLALDestroyREAL4FrequencySeries");
	XLALDestroyREAL4FrequencySeries (series);
	RETURN(status);
}


REAL4FrequencySeries *XLALCreateREAL4FrequencySeries (
	const CHAR *name,
	const LIGOTimeGPS *epoch,
	REAL8 f0,
	REAL8 deltaF,
	const LALUnit *sampleUnits,
	size_t length
)
{
	static const char func[] = "XLALCreateREAL4FrequencySeries";
	REAL4FrequencySeries *new;
	REAL4Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCreateREAL4Sequence (length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyREAL4Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	if(name)
		strncpy(new->name, name, LALNameLength);
	else
		new->name[0] = '\0';
	new->epoch = *epoch;
	new->f0 = f0;
	new->deltaF = deltaF;
	new->sampleUnits = *sampleUnits;
	new->data = sequence;

	return new;
}


void LALCreateREAL4FrequencySeries (
	LALStatus *status,
	REAL4FrequencySeries **output,
	const CHAR *name,
	LIGOTimeGPS epoch,
	REAL8 f0,
	REAL8 deltaF,
	LALUnit sampleUnits,
	size_t length
)
{
	INITSTATUS(status, "LALCreateREAL4FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALCreateREAL4FrequencySeries", "XLALCreateREAL4FrequencySeries");
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateREAL4FrequencySeries (name, &epoch, f0, deltaF, &sampleUnits, length);
	if(*output == NULL) {
		XLALClearErrno();
		ABORT(status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	}
	RETURN(status);
}


REAL4FrequencySeries *XLALCutREAL4FrequencySeries (
	const REAL4FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALCutREAL4FrequencySeries";
	REAL4FrequencySeries *new;
	REAL4Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCutREAL4Sequence (series->data, first, length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyREAL4Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	*new = *series;
	new->data = sequence;
	new->f0 += first * new->deltaF;

	return new;
}


REAL4FrequencySeries *XLALResizeREAL4FrequencySeries (
	REAL4FrequencySeries *series,
	int first,
	size_t length
)
{
	static const char func[] = "XLALResizeREAL4FrequencySeries";

	if(!XLALResizeREAL4Sequence (series->data, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	series->f0 += first * series->deltaF;

	return series;
}


REAL4FrequencySeries *XLALShrinkREAL4FrequencySeries (
	REAL4FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALShrinkREAL4FrequencySeries";

	if(!XLALResizeREAL4FrequencySeries (series, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);

	return series;
}


REAL4FrequencySeries *XLALAddREAL4FrequencySeries (
	REAL4FrequencySeries *arg1,
	const REAL4FrequencySeries *arg2
)
{
	static const char func[] = "XLALAddREAL4FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */
	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */
	if(Delta_f0 >= 0) {
		i = floor(Delta_f0 / arg1->deltaF + 0.5);
		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* add arg2 to arg1, adjusting the units */
	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		arg1->data->data[i] += arg2->data->data[j] * unit_ratio;
	}

	return arg1;
}





REAL4FrequencySeries *XLALMultiplyREAL4FrequencySeries (
	REAL4FrequencySeries *arg1,
	const REAL4FrequencySeries *arg2
)
{
	static const char func[] = "XLALMultiplyREAL4FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */

	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */

	if(Delta_f0 >= 0) {
		unsigned start = floor(Delta_f0 / arg1->deltaF + 0.5);

		/* zero the first part of arg1 */

		for(i = 0; i < start && i < arg1->data->length; i++) {
			arg1->data->data[i] = 0.0;
		}

		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* multiply arg1 by arg2 */

	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		arg1->data->data[i] *= arg2->data->data[j] * unit_ratio;
	}

	/* zero the last part of arg1 */

	for(; i < arg1->data->length; i++) {
		arg1->data->data[i] = 0.0;
	}

	return arg1;
}






void XLALDestroyREAL8FrequencySeries (
	REAL8FrequencySeries *series
)
{
	if(series)
		XLALDestroyREAL8Sequence (series->data);
	XLALFree(series);
}


void LALDestroyREAL8FrequencySeries (
	LALStatus *status,
	REAL8FrequencySeries *series
)
{
	INITSTATUS(status, "LALDestroyREAL8FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALDestroyREAL8FrequencySeries", "XLALDestroyREAL8FrequencySeries");
	XLALDestroyREAL8FrequencySeries (series);
	RETURN(status);
}


REAL8FrequencySeries *XLALCreateREAL8FrequencySeries (
	const CHAR *name,
	const LIGOTimeGPS *epoch,
	REAL8 f0,
	REAL8 deltaF,
	const LALUnit *sampleUnits,
	size_t length
)
{
	static const char func[] = "XLALCreateREAL8FrequencySeries";
	REAL8FrequencySeries *new;
	REAL8Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCreateREAL8Sequence (length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyREAL8Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	if(name)
		strncpy(new->name, name, LALNameLength);
	else
		new->name[0] = '\0';
	new->epoch = *epoch;
	new->f0 = f0;
	new->deltaF = deltaF;
	new->sampleUnits = *sampleUnits;
	new->data = sequence;

	return new;
}


void LALCreateREAL8FrequencySeries (
	LALStatus *status,
	REAL8FrequencySeries **output,
	const CHAR *name,
	LIGOTimeGPS epoch,
	REAL8 f0,
	REAL8 deltaF,
	LALUnit sampleUnits,
	size_t length
)
{
	INITSTATUS(status, "LALCreateREAL8FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALCreateREAL8FrequencySeries", "XLALCreateREAL8FrequencySeries");
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateREAL8FrequencySeries (name, &epoch, f0, deltaF, &sampleUnits, length);
	if(*output == NULL) {
		XLALClearErrno();
		ABORT(status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	}
	RETURN(status);
}


REAL8FrequencySeries *XLALCutREAL8FrequencySeries (
	const REAL8FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALCutREAL8FrequencySeries";
	REAL8FrequencySeries *new;
	REAL8Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCutREAL8Sequence (series->data, first, length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyREAL8Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	*new = *series;
	new->data = sequence;
	new->f0 += first * new->deltaF;

	return new;
}


REAL8FrequencySeries *XLALResizeREAL8FrequencySeries (
	REAL8FrequencySeries *series,
	int first,
	size_t length
)
{
	static const char func[] = "XLALResizeREAL8FrequencySeries";

	if(!XLALResizeREAL8Sequence (series->data, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	series->f0 += first * series->deltaF;

	return series;
}


REAL8FrequencySeries *XLALShrinkREAL8FrequencySeries (
	REAL8FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALShrinkREAL8FrequencySeries";

	if(!XLALResizeREAL8FrequencySeries (series, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);

	return series;
}


REAL8FrequencySeries *XLALAddREAL8FrequencySeries (
	REAL8FrequencySeries *arg1,
	const REAL8FrequencySeries *arg2
)
{
	static const char func[] = "XLALAddREAL8FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */
	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */
	if(Delta_f0 >= 0) {
		i = floor(Delta_f0 / arg1->deltaF + 0.5);
		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* add arg2 to arg1, adjusting the units */
	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		arg1->data->data[i] += arg2->data->data[j] * unit_ratio;
	}

	return arg1;
}





REAL8FrequencySeries *XLALMultiplyREAL8FrequencySeries (
	REAL8FrequencySeries *arg1,
	const REAL8FrequencySeries *arg2
)
{
	static const char func[] = "XLALMultiplyREAL8FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */

	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */

	if(Delta_f0 >= 0) {
		unsigned start = floor(Delta_f0 / arg1->deltaF + 0.5);

		/* zero the first part of arg1 */

		for(i = 0; i < start && i < arg1->data->length; i++) {
			arg1->data->data[i] = 0.0;
		}

		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* multiply arg1 by arg2 */

	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		arg1->data->data[i] *= arg2->data->data[j] * unit_ratio;
	}

	/* zero the last part of arg1 */

	for(; i < arg1->data->length; i++) {
		arg1->data->data[i] = 0.0;
	}

	return arg1;
}






void XLALDestroyCOMPLEX8FrequencySeries (
	COMPLEX8FrequencySeries *series
)
{
	if(series)
		XLALDestroyCOMPLEX8Sequence (series->data);
	XLALFree(series);
}


void LALDestroyCOMPLEX8FrequencySeries (
	LALStatus *status,
	COMPLEX8FrequencySeries *series
)
{
	INITSTATUS(status, "LALDestroyCOMPLEX8FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALDestroyCOMPLEX8FrequencySeries", "XLALDestroyCOMPLEX8FrequencySeries");
	XLALDestroyCOMPLEX8FrequencySeries (series);
	RETURN(status);
}


COMPLEX8FrequencySeries *XLALCreateCOMPLEX8FrequencySeries (
	const CHAR *name,
	const LIGOTimeGPS *epoch,
	REAL8 f0,
	REAL8 deltaF,
	const LALUnit *sampleUnits,
	size_t length
)
{
	static const char func[] = "XLALCreateCOMPLEX8FrequencySeries";
	COMPLEX8FrequencySeries *new;
	COMPLEX8Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCreateCOMPLEX8Sequence (length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyCOMPLEX8Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	if(name)
		strncpy(new->name, name, LALNameLength);
	else
		new->name[0] = '\0';
	new->epoch = *epoch;
	new->f0 = f0;
	new->deltaF = deltaF;
	new->sampleUnits = *sampleUnits;
	new->data = sequence;

	return new;
}


void LALCreateCOMPLEX8FrequencySeries (
	LALStatus *status,
	COMPLEX8FrequencySeries **output,
	const CHAR *name,
	LIGOTimeGPS epoch,
	REAL8 f0,
	REAL8 deltaF,
	LALUnit sampleUnits,
	size_t length
)
{
	INITSTATUS(status, "LALCreateCOMPLEX8FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALCreateCOMPLEX8FrequencySeries", "XLALCreateCOMPLEX8FrequencySeries");
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateCOMPLEX8FrequencySeries (name, &epoch, f0, deltaF, &sampleUnits, length);
	if(*output == NULL) {
		XLALClearErrno();
		ABORT(status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	}
	RETURN(status);
}


COMPLEX8FrequencySeries *XLALCutCOMPLEX8FrequencySeries (
	const COMPLEX8FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALCutCOMPLEX8FrequencySeries";
	COMPLEX8FrequencySeries *new;
	COMPLEX8Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCutCOMPLEX8Sequence (series->data, first, length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyCOMPLEX8Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	*new = *series;
	new->data = sequence;
	new->f0 += first * new->deltaF;

	return new;
}


COMPLEX8FrequencySeries *XLALResizeCOMPLEX8FrequencySeries (
	COMPLEX8FrequencySeries *series,
	int first,
	size_t length
)
{
	static const char func[] = "XLALResizeCOMPLEX8FrequencySeries";

	if(!XLALResizeCOMPLEX8Sequence (series->data, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	series->f0 += first * series->deltaF;

	return series;
}


COMPLEX8FrequencySeries *XLALShrinkCOMPLEX8FrequencySeries (
	COMPLEX8FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALShrinkCOMPLEX8FrequencySeries";

	if(!XLALResizeCOMPLEX8FrequencySeries (series, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);

	return series;
}


COMPLEX8FrequencySeries *XLALAddCOMPLEX8FrequencySeries (
	COMPLEX8FrequencySeries *arg1,
	const COMPLEX8FrequencySeries *arg2
)
{
	static const char func[] = "XLALAddCOMPLEX8FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */
	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */
	if(Delta_f0 >= 0) {
		i = floor(Delta_f0 / arg1->deltaF + 0.5);
		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* add arg2 to arg1, adjusting the units */
	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		arg1->data->data[i].re += arg2->data->data[j].re * unit_ratio;
		arg1->data->data[i].im += arg2->data->data[j].im * unit_ratio;
		
	}

	return arg1;
}


COMPLEX8FrequencySeries *XLALConjugateCOMPLEX8FrequencySeries (
	COMPLEX8FrequencySeries *series
)
{
	XLALConjugateCOMPLEX8Sequence (series->data);
	return series;

}



COMPLEX8FrequencySeries *XLALMultiplyCOMPLEX8FrequencySeries (
	COMPLEX8FrequencySeries *arg1,
	const COMPLEX8FrequencySeries *arg2
)
{
	static const char func[] = "XLALMultiplyCOMPLEX8FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */

	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */

	if(Delta_f0 >= 0) {
		unsigned start = floor(Delta_f0 / arg1->deltaF + 0.5);

		/* zero the first part of arg1 */

		for(i = 0; i < start && i < arg1->data->length; i++) {
			arg1->data->data[i] = LAL_COMPLEX8_ZERO;
			
		}

		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* multiply arg1 by arg2 */

	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		REAL4 re = arg2->data->data[j].re * unit_ratio;
		REAL4 im = arg2->data->data[j].im * unit_ratio;
		arg1->data->data[i].re = arg1->data->data[i].re * re - arg1->data->data[i].im * im;
		arg1->data->data[i].im = arg1->data->data[i].re * im + arg1->data->data[i].im * re;
		
	}

	/* zero the last part of arg1 */

	for(; i < arg1->data->length; i++) {
		arg1->data->data[i] = LAL_COMPLEX8_ZERO;
		
	}

	return arg1;
}






void XLALDestroyCOMPLEX16FrequencySeries (
	COMPLEX16FrequencySeries *series
)
{
	if(series)
		XLALDestroyCOMPLEX16Sequence (series->data);
	XLALFree(series);
}


void LALDestroyCOMPLEX16FrequencySeries (
	LALStatus *status,
	COMPLEX16FrequencySeries *series
)
{
	INITSTATUS(status, "LALDestroyCOMPLEX16FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALDestroyCOMPLEX16FrequencySeries", "XLALDestroyCOMPLEX16FrequencySeries");
	XLALDestroyCOMPLEX16FrequencySeries (series);
	RETURN(status);
}


COMPLEX16FrequencySeries *XLALCreateCOMPLEX16FrequencySeries (
	const CHAR *name,
	const LIGOTimeGPS *epoch,
	REAL8 f0,
	REAL8 deltaF,
	const LALUnit *sampleUnits,
	size_t length
)
{
	static const char func[] = "XLALCreateCOMPLEX16FrequencySeries";
	COMPLEX16FrequencySeries *new;
	COMPLEX16Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCreateCOMPLEX16Sequence (length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyCOMPLEX16Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	if(name)
		strncpy(new->name, name, LALNameLength);
	else
		new->name[0] = '\0';
	new->epoch = *epoch;
	new->f0 = f0;
	new->deltaF = deltaF;
	new->sampleUnits = *sampleUnits;
	new->data = sequence;

	return new;
}


void LALCreateCOMPLEX16FrequencySeries (
	LALStatus *status,
	COMPLEX16FrequencySeries **output,
	const CHAR *name,
	LIGOTimeGPS epoch,
	REAL8 f0,
	REAL8 deltaF,
	LALUnit sampleUnits,
	size_t length
)
{
	INITSTATUS(status, "LALCreateCOMPLEX16FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALCreateCOMPLEX16FrequencySeries", "XLALCreateCOMPLEX16FrequencySeries");
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateCOMPLEX16FrequencySeries (name, &epoch, f0, deltaF, &sampleUnits, length);
	if(*output == NULL) {
		XLALClearErrno();
		ABORT(status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	}
	RETURN(status);
}


COMPLEX16FrequencySeries *XLALCutCOMPLEX16FrequencySeries (
	const COMPLEX16FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALCutCOMPLEX16FrequencySeries";
	COMPLEX16FrequencySeries *new;
	COMPLEX16Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCutCOMPLEX16Sequence (series->data, first, length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyCOMPLEX16Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	*new = *series;
	new->data = sequence;
	new->f0 += first * new->deltaF;

	return new;
}


COMPLEX16FrequencySeries *XLALResizeCOMPLEX16FrequencySeries (
	COMPLEX16FrequencySeries *series,
	int first,
	size_t length
)
{
	static const char func[] = "XLALResizeCOMPLEX16FrequencySeries";

	if(!XLALResizeCOMPLEX16Sequence (series->data, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	series->f0 += first * series->deltaF;

	return series;
}


COMPLEX16FrequencySeries *XLALShrinkCOMPLEX16FrequencySeries (
	COMPLEX16FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALShrinkCOMPLEX16FrequencySeries";

	if(!XLALResizeCOMPLEX16FrequencySeries (series, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);

	return series;
}


COMPLEX16FrequencySeries *XLALAddCOMPLEX16FrequencySeries (
	COMPLEX16FrequencySeries *arg1,
	const COMPLEX16FrequencySeries *arg2
)
{
	static const char func[] = "XLALAddCOMPLEX16FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */
	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */
	if(Delta_f0 >= 0) {
		i = floor(Delta_f0 / arg1->deltaF + 0.5);
		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* add arg2 to arg1, adjusting the units */
	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		arg1->data->data[i].re += arg2->data->data[j].re * unit_ratio;
		arg1->data->data[i].im += arg2->data->data[j].im * unit_ratio;
		
	}

	return arg1;
}


COMPLEX16FrequencySeries *XLALConjugateCOMPLEX16FrequencySeries (
	COMPLEX16FrequencySeries *series
)
{
	XLALConjugateCOMPLEX16Sequence (series->data);
	return series;
}



COMPLEX16FrequencySeries *XLALMultiplyCOMPLEX16FrequencySeries (
	COMPLEX16FrequencySeries *arg1,
	const COMPLEX16FrequencySeries *arg2
)
{
	static const char func[] = "XLALMultiplyCOMPLEX16FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */

	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */

	if(Delta_f0 >= 0) {
		unsigned start = floor(Delta_f0 / arg1->deltaF + 0.5);

		/* zero the first part of arg1 */

		for(i = 0; i < start && i < arg1->data->length; i++) {
			arg1->data->data[i] = LAL_COMPLEX16_ZERO;
			
		}

		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* multiply arg1 by arg2 */

	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		REAL8 re = arg2->data->data[j].re * unit_ratio;
		REAL8 im = arg2->data->data[j].im * unit_ratio;
		arg1->data->data[i].re = arg1->data->data[i].re * re - arg1->data->data[i].im * im;
		arg1->data->data[i].im = arg1->data->data[i].re * im + arg1->data->data[i].im * re;
		
	}

	/* zero the last part of arg1 */

	for(; i < arg1->data->length; i++) {
		arg1->data->data[i] = LAL_COMPLEX16_ZERO;
		
	}

	return arg1;
}






void XLALDestroyINT2FrequencySeries (
	INT2FrequencySeries *series
)
{
	if(series)
		XLALDestroyINT2Sequence (series->data);
	XLALFree(series);
}


void LALDestroyINT2FrequencySeries (
	LALStatus *status,
	INT2FrequencySeries *series
)
{
	INITSTATUS(status, "LALDestroyINT2FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALDestroyINT2FrequencySeries", "XLALDestroyINT2FrequencySeries");
	XLALDestroyINT2FrequencySeries (series);
	RETURN(status);
}


INT2FrequencySeries *XLALCreateINT2FrequencySeries (
	const CHAR *name,
	const LIGOTimeGPS *epoch,
	REAL8 f0,
	REAL8 deltaF,
	const LALUnit *sampleUnits,
	size_t length
)
{
	static const char func[] = "XLALCreateINT2FrequencySeries";
	INT2FrequencySeries *new;
	INT2Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCreateINT2Sequence (length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyINT2Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	if(name)
		strncpy(new->name, name, LALNameLength);
	else
		new->name[0] = '\0';
	new->epoch = *epoch;
	new->f0 = f0;
	new->deltaF = deltaF;
	new->sampleUnits = *sampleUnits;
	new->data = sequence;

	return new;
}


void LALCreateINT2FrequencySeries (
	LALStatus *status,
	INT2FrequencySeries **output,
	const CHAR *name,
	LIGOTimeGPS epoch,
	REAL8 f0,
	REAL8 deltaF,
	LALUnit sampleUnits,
	size_t length
)
{
	INITSTATUS(status, "LALCreateINT2FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALCreateINT2FrequencySeries", "XLALCreateINT2FrequencySeries");
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateINT2FrequencySeries (name, &epoch, f0, deltaF, &sampleUnits, length);
	if(*output == NULL) {
		XLALClearErrno();
		ABORT(status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	}
	RETURN(status);
}


INT2FrequencySeries *XLALCutINT2FrequencySeries (
	const INT2FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALCutINT2FrequencySeries";
	INT2FrequencySeries *new;
	INT2Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCutINT2Sequence (series->data, first, length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyINT2Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	*new = *series;
	new->data = sequence;
	new->f0 += first * new->deltaF;

	return new;
}


INT2FrequencySeries *XLALResizeINT2FrequencySeries (
	INT2FrequencySeries *series,
	int first,
	size_t length
)
{
	static const char func[] = "XLALResizeINT2FrequencySeries";

	if(!XLALResizeINT2Sequence (series->data, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	series->f0 += first * series->deltaF;

	return series;
}


INT2FrequencySeries *XLALShrinkINT2FrequencySeries (
	INT2FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALShrinkINT2FrequencySeries";

	if(!XLALResizeINT2FrequencySeries (series, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);

	return series;
}


INT2FrequencySeries *XLALAddINT2FrequencySeries (
	INT2FrequencySeries *arg1,
	const INT2FrequencySeries *arg2
)
{
	static const char func[] = "XLALAddINT2FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */
	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */
	if(Delta_f0 >= 0) {
		i = floor(Delta_f0 / arg1->deltaF + 0.5);
		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* add arg2 to arg1, adjusting the units */
	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		arg1->data->data[i] += arg2->data->data[j] * unit_ratio;
	}

	return arg1;
}





INT2FrequencySeries *XLALMultiplyINT2FrequencySeries (
	INT2FrequencySeries *arg1,
	const INT2FrequencySeries *arg2
)
{
	static const char func[] = "XLALMultiplyINT2FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */

	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */

	if(Delta_f0 >= 0) {
		unsigned start = floor(Delta_f0 / arg1->deltaF + 0.5);

		/* zero the first part of arg1 */

		for(i = 0; i < start && i < arg1->data->length; i++) {
			arg1->data->data[i] = 0.0;
		}

		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* multiply arg1 by arg2 */

	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		arg1->data->data[i] *= arg2->data->data[j] * unit_ratio;
	}

	/* zero the last part of arg1 */

	for(; i < arg1->data->length; i++) {
		arg1->data->data[i] = 0.0;
	}

	return arg1;
}






void XLALDestroyUINT2FrequencySeries (
	UINT2FrequencySeries *series
)
{
	if(series)
		XLALDestroyUINT2Sequence (series->data);
	XLALFree(series);
}


void LALDestroyUINT2FrequencySeries (
	LALStatus *status,
	UINT2FrequencySeries *series
)
{
	INITSTATUS(status, "LALDestroyUINT2FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALDestroyUINT2FrequencySeries", "XLALDestroyUINT2FrequencySeries");
	XLALDestroyUINT2FrequencySeries (series);
	RETURN(status);
}


UINT2FrequencySeries *XLALCreateUINT2FrequencySeries (
	const CHAR *name,
	const LIGOTimeGPS *epoch,
	REAL8 f0,
	REAL8 deltaF,
	const LALUnit *sampleUnits,
	size_t length
)
{
	static const char func[] = "XLALCreateUINT2FrequencySeries";
	UINT2FrequencySeries *new;
	UINT2Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCreateUINT2Sequence (length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyUINT2Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	if(name)
		strncpy(new->name, name, LALNameLength);
	else
		new->name[0] = '\0';
	new->epoch = *epoch;
	new->f0 = f0;
	new->deltaF = deltaF;
	new->sampleUnits = *sampleUnits;
	new->data = sequence;

	return new;
}


void LALCreateUINT2FrequencySeries (
	LALStatus *status,
	UINT2FrequencySeries **output,
	const CHAR *name,
	LIGOTimeGPS epoch,
	REAL8 f0,
	REAL8 deltaF,
	LALUnit sampleUnits,
	size_t length
)
{
	INITSTATUS(status, "LALCreateUINT2FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALCreateUINT2FrequencySeries", "XLALCreateUINT2FrequencySeries");
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateUINT2FrequencySeries (name, &epoch, f0, deltaF, &sampleUnits, length);
	if(*output == NULL) {
		XLALClearErrno();
		ABORT(status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	}
	RETURN(status);
}


UINT2FrequencySeries *XLALCutUINT2FrequencySeries (
	const UINT2FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALCutUINT2FrequencySeries";
	UINT2FrequencySeries *new;
	UINT2Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCutUINT2Sequence (series->data, first, length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyUINT2Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	*new = *series;
	new->data = sequence;
	new->f0 += first * new->deltaF;

	return new;
}


UINT2FrequencySeries *XLALResizeUINT2FrequencySeries (
	UINT2FrequencySeries *series,
	int first,
	size_t length
)
{
	static const char func[] = "XLALResizeUINT2FrequencySeries";

	if(!XLALResizeUINT2Sequence (series->data, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	series->f0 += first * series->deltaF;

	return series;
}


UINT2FrequencySeries *XLALShrinkUINT2FrequencySeries (
	UINT2FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALShrinkUINT2FrequencySeries";

	if(!XLALResizeUINT2FrequencySeries (series, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);

	return series;
}


UINT2FrequencySeries *XLALAddUINT2FrequencySeries (
	UINT2FrequencySeries *arg1,
	const UINT2FrequencySeries *arg2
)
{
	static const char func[] = "XLALAddUINT2FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */
	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */
	if(Delta_f0 >= 0) {
		i = floor(Delta_f0 / arg1->deltaF + 0.5);
		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* add arg2 to arg1, adjusting the units */
	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		arg1->data->data[i] += arg2->data->data[j] * unit_ratio;
	}

	return arg1;
}





UINT2FrequencySeries *XLALMultiplyUINT2FrequencySeries (
	UINT2FrequencySeries *arg1,
	const UINT2FrequencySeries *arg2
)
{
	static const char func[] = "XLALMultiplyUINT2FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */

	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */

	if(Delta_f0 >= 0) {
		unsigned start = floor(Delta_f0 / arg1->deltaF + 0.5);

		/* zero the first part of arg1 */

		for(i = 0; i < start && i < arg1->data->length; i++) {
			arg1->data->data[i] = 0.0;
		}

		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* multiply arg1 by arg2 */

	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		arg1->data->data[i] *= arg2->data->data[j] * unit_ratio;
	}

	/* zero the last part of arg1 */

	for(; i < arg1->data->length; i++) {
		arg1->data->data[i] = 0.0;
	}

	return arg1;
}






void XLALDestroyINT4FrequencySeries (
	INT4FrequencySeries *series
)
{
	if(series)
		XLALDestroyINT4Sequence (series->data);
	XLALFree(series);
}


void LALDestroyINT4FrequencySeries (
	LALStatus *status,
	INT4FrequencySeries *series
)
{
	INITSTATUS(status, "LALDestroyINT4FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALDestroyINT4FrequencySeries", "XLALDestroyINT4FrequencySeries");
	XLALDestroyINT4FrequencySeries (series);
	RETURN(status);
}


INT4FrequencySeries *XLALCreateINT4FrequencySeries (
	const CHAR *name,
	const LIGOTimeGPS *epoch,
	REAL8 f0,
	REAL8 deltaF,
	const LALUnit *sampleUnits,
	size_t length
)
{
	static const char func[] = "XLALCreateINT4FrequencySeries";
	INT4FrequencySeries *new;
	INT4Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCreateINT4Sequence (length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyINT4Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	if(name)
		strncpy(new->name, name, LALNameLength);
	else
		new->name[0] = '\0';
	new->epoch = *epoch;
	new->f0 = f0;
	new->deltaF = deltaF;
	new->sampleUnits = *sampleUnits;
	new->data = sequence;

	return new;
}


void LALCreateINT4FrequencySeries (
	LALStatus *status,
	INT4FrequencySeries **output,
	const CHAR *name,
	LIGOTimeGPS epoch,
	REAL8 f0,
	REAL8 deltaF,
	LALUnit sampleUnits,
	size_t length
)
{
	INITSTATUS(status, "LALCreateINT4FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALCreateINT4FrequencySeries", "XLALCreateINT4FrequencySeries");
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateINT4FrequencySeries (name, &epoch, f0, deltaF, &sampleUnits, length);
	if(*output == NULL) {
		XLALClearErrno();
		ABORT(status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	}
	RETURN(status);
}


INT4FrequencySeries *XLALCutINT4FrequencySeries (
	const INT4FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALCutINT4FrequencySeries";
	INT4FrequencySeries *new;
	INT4Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCutINT4Sequence (series->data, first, length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyINT4Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	*new = *series;
	new->data = sequence;
	new->f0 += first * new->deltaF;

	return new;
}


INT4FrequencySeries *XLALResizeINT4FrequencySeries (
	INT4FrequencySeries *series,
	int first,
	size_t length
)
{
	static const char func[] = "XLALResizeINT4FrequencySeries";

	if(!XLALResizeINT4Sequence (series->data, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	series->f0 += first * series->deltaF;

	return series;
}


INT4FrequencySeries *XLALShrinkINT4FrequencySeries (
	INT4FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALShrinkINT4FrequencySeries";

	if(!XLALResizeINT4FrequencySeries (series, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);

	return series;
}


INT4FrequencySeries *XLALAddINT4FrequencySeries (
	INT4FrequencySeries *arg1,
	const INT4FrequencySeries *arg2
)
{
	static const char func[] = "XLALAddINT4FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */
	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */
	if(Delta_f0 >= 0) {
		i = floor(Delta_f0 / arg1->deltaF + 0.5);
		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* add arg2 to arg1, adjusting the units */
	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		arg1->data->data[i] += arg2->data->data[j] * unit_ratio;
	}

	return arg1;
}





INT4FrequencySeries *XLALMultiplyINT4FrequencySeries (
	INT4FrequencySeries *arg1,
	const INT4FrequencySeries *arg2
)
{
	static const char func[] = "XLALMultiplyINT4FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */

	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */

	if(Delta_f0 >= 0) {
		unsigned start = floor(Delta_f0 / arg1->deltaF + 0.5);

		/* zero the first part of arg1 */

		for(i = 0; i < start && i < arg1->data->length; i++) {
			arg1->data->data[i] = 0.0;
		}

		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* multiply arg1 by arg2 */

	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		arg1->data->data[i] *= arg2->data->data[j] * unit_ratio;
	}

	/* zero the last part of arg1 */

	for(; i < arg1->data->length; i++) {
		arg1->data->data[i] = 0.0;
	}

	return arg1;
}






void XLALDestroyUINT4FrequencySeries (
	UINT4FrequencySeries *series
)
{
	if(series)
		XLALDestroyUINT4Sequence (series->data);
	XLALFree(series);
}


void LALDestroyUINT4FrequencySeries (
	LALStatus *status,
	UINT4FrequencySeries *series
)
{
	INITSTATUS(status, "LALDestroyUINT4FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALDestroyUINT4FrequencySeries", "XLALDestroyUINT4FrequencySeries");
	XLALDestroyUINT4FrequencySeries (series);
	RETURN(status);
}


UINT4FrequencySeries *XLALCreateUINT4FrequencySeries (
	const CHAR *name,
	const LIGOTimeGPS *epoch,
	REAL8 f0,
	REAL8 deltaF,
	const LALUnit *sampleUnits,
	size_t length
)
{
	static const char func[] = "XLALCreateUINT4FrequencySeries";
	UINT4FrequencySeries *new;
	UINT4Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCreateUINT4Sequence (length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyUINT4Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	if(name)
		strncpy(new->name, name, LALNameLength);
	else
		new->name[0] = '\0';
	new->epoch = *epoch;
	new->f0 = f0;
	new->deltaF = deltaF;
	new->sampleUnits = *sampleUnits;
	new->data = sequence;

	return new;
}


void LALCreateUINT4FrequencySeries (
	LALStatus *status,
	UINT4FrequencySeries **output,
	const CHAR *name,
	LIGOTimeGPS epoch,
	REAL8 f0,
	REAL8 deltaF,
	LALUnit sampleUnits,
	size_t length
)
{
	INITSTATUS(status, "LALCreateUINT4FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALCreateUINT4FrequencySeries", "XLALCreateUINT4FrequencySeries");
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateUINT4FrequencySeries (name, &epoch, f0, deltaF, &sampleUnits, length);
	if(*output == NULL) {
		XLALClearErrno();
		ABORT(status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	}
	RETURN(status);
}


UINT4FrequencySeries *XLALCutUINT4FrequencySeries (
	const UINT4FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALCutUINT4FrequencySeries";
	UINT4FrequencySeries *new;
	UINT4Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCutUINT4Sequence (series->data, first, length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyUINT4Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	*new = *series;
	new->data = sequence;
	new->f0 += first * new->deltaF;

	return new;
}


UINT4FrequencySeries *XLALResizeUINT4FrequencySeries (
	UINT4FrequencySeries *series,
	int first,
	size_t length
)
{
	static const char func[] = "XLALResizeUINT4FrequencySeries";

	if(!XLALResizeUINT4Sequence (series->data, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	series->f0 += first * series->deltaF;

	return series;
}


UINT4FrequencySeries *XLALShrinkUINT4FrequencySeries (
	UINT4FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALShrinkUINT4FrequencySeries";

	if(!XLALResizeUINT4FrequencySeries (series, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);

	return series;
}


UINT4FrequencySeries *XLALAddUINT4FrequencySeries (
	UINT4FrequencySeries *arg1,
	const UINT4FrequencySeries *arg2
)
{
	static const char func[] = "XLALAddUINT4FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */
	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */
	if(Delta_f0 >= 0) {
		i = floor(Delta_f0 / arg1->deltaF + 0.5);
		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* add arg2 to arg1, adjusting the units */
	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		arg1->data->data[i] += arg2->data->data[j] * unit_ratio;
	}

	return arg1;
}





UINT4FrequencySeries *XLALMultiplyUINT4FrequencySeries (
	UINT4FrequencySeries *arg1,
	const UINT4FrequencySeries *arg2
)
{
	static const char func[] = "XLALMultiplyUINT4FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */

	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */

	if(Delta_f0 >= 0) {
		unsigned start = floor(Delta_f0 / arg1->deltaF + 0.5);

		/* zero the first part of arg1 */

		for(i = 0; i < start && i < arg1->data->length; i++) {
			arg1->data->data[i] = 0.0;
		}

		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* multiply arg1 by arg2 */

	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		arg1->data->data[i] *= arg2->data->data[j] * unit_ratio;
	}

	/* zero the last part of arg1 */

	for(; i < arg1->data->length; i++) {
		arg1->data->data[i] = 0.0;
	}

	return arg1;
}






void XLALDestroyINT8FrequencySeries (
	INT8FrequencySeries *series
)
{
	if(series)
		XLALDestroyINT8Sequence (series->data);
	XLALFree(series);
}


void LALDestroyINT8FrequencySeries (
	LALStatus *status,
	INT8FrequencySeries *series
)
{
	INITSTATUS(status, "LALDestroyINT8FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALDestroyINT8FrequencySeries", "XLALDestroyINT8FrequencySeries");
	XLALDestroyINT8FrequencySeries (series);
	RETURN(status);
}


INT8FrequencySeries *XLALCreateINT8FrequencySeries (
	const CHAR *name,
	const LIGOTimeGPS *epoch,
	REAL8 f0,
	REAL8 deltaF,
	const LALUnit *sampleUnits,
	size_t length
)
{
	static const char func[] = "XLALCreateINT8FrequencySeries";
	INT8FrequencySeries *new;
	INT8Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCreateINT8Sequence (length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyINT8Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	if(name)
		strncpy(new->name, name, LALNameLength);
	else
		new->name[0] = '\0';
	new->epoch = *epoch;
	new->f0 = f0;
	new->deltaF = deltaF;
	new->sampleUnits = *sampleUnits;
	new->data = sequence;

	return new;
}


void LALCreateINT8FrequencySeries (
	LALStatus *status,
	INT8FrequencySeries **output,
	const CHAR *name,
	LIGOTimeGPS epoch,
	REAL8 f0,
	REAL8 deltaF,
	LALUnit sampleUnits,
	size_t length
)
{
	INITSTATUS(status, "LALCreateINT8FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALCreateINT8FrequencySeries", "XLALCreateINT8FrequencySeries");
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateINT8FrequencySeries (name, &epoch, f0, deltaF, &sampleUnits, length);
	if(*output == NULL) {
		XLALClearErrno();
		ABORT(status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	}
	RETURN(status);
}


INT8FrequencySeries *XLALCutINT8FrequencySeries (
	const INT8FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALCutINT8FrequencySeries";
	INT8FrequencySeries *new;
	INT8Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCutINT8Sequence (series->data, first, length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyINT8Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	*new = *series;
	new->data = sequence;
	new->f0 += first * new->deltaF;

	return new;
}


INT8FrequencySeries *XLALResizeINT8FrequencySeries (
	INT8FrequencySeries *series,
	int first,
	size_t length
)
{
	static const char func[] = "XLALResizeINT8FrequencySeries";

	if(!XLALResizeINT8Sequence (series->data, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	series->f0 += first * series->deltaF;

	return series;
}


INT8FrequencySeries *XLALShrinkINT8FrequencySeries (
	INT8FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALShrinkINT8FrequencySeries";

	if(!XLALResizeINT8FrequencySeries (series, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);

	return series;
}


INT8FrequencySeries *XLALAddINT8FrequencySeries (
	INT8FrequencySeries *arg1,
	const INT8FrequencySeries *arg2
)
{
	static const char func[] = "XLALAddINT8FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */
	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */
	if(Delta_f0 >= 0) {
		i = floor(Delta_f0 / arg1->deltaF + 0.5);
		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* add arg2 to arg1, adjusting the units */
	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		arg1->data->data[i] += arg2->data->data[j] * unit_ratio;
	}

	return arg1;
}





INT8FrequencySeries *XLALMultiplyINT8FrequencySeries (
	INT8FrequencySeries *arg1,
	const INT8FrequencySeries *arg2
)
{
	static const char func[] = "XLALMultiplyINT8FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */

	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */

	if(Delta_f0 >= 0) {
		unsigned start = floor(Delta_f0 / arg1->deltaF + 0.5);

		/* zero the first part of arg1 */

		for(i = 0; i < start && i < arg1->data->length; i++) {
			arg1->data->data[i] = 0.0;
		}

		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* multiply arg1 by arg2 */

	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		arg1->data->data[i] *= arg2->data->data[j] * unit_ratio;
	}

	/* zero the last part of arg1 */

	for(; i < arg1->data->length; i++) {
		arg1->data->data[i] = 0.0;
	}

	return arg1;
}






void XLALDestroyUINT8FrequencySeries (
	UINT8FrequencySeries *series
)
{
	if(series)
		XLALDestroyUINT8Sequence (series->data);
	XLALFree(series);
}


void LALDestroyUINT8FrequencySeries (
	LALStatus *status,
	UINT8FrequencySeries *series
)
{
	INITSTATUS(status, "LALDestroyUINT8FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALDestroyUINT8FrequencySeries", "XLALDestroyUINT8FrequencySeries");
	XLALDestroyUINT8FrequencySeries (series);
	RETURN(status);
}


UINT8FrequencySeries *XLALCreateUINT8FrequencySeries (
	const CHAR *name,
	const LIGOTimeGPS *epoch,
	REAL8 f0,
	REAL8 deltaF,
	const LALUnit *sampleUnits,
	size_t length
)
{
	static const char func[] = "XLALCreateUINT8FrequencySeries";
	UINT8FrequencySeries *new;
	UINT8Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCreateUINT8Sequence (length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyUINT8Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	if(name)
		strncpy(new->name, name, LALNameLength);
	else
		new->name[0] = '\0';
	new->epoch = *epoch;
	new->f0 = f0;
	new->deltaF = deltaF;
	new->sampleUnits = *sampleUnits;
	new->data = sequence;

	return new;
}


void LALCreateUINT8FrequencySeries (
	LALStatus *status,
	UINT8FrequencySeries **output,
	const CHAR *name,
	LIGOTimeGPS epoch,
	REAL8 f0,
	REAL8 deltaF,
	LALUnit sampleUnits,
	size_t length
)
{
	INITSTATUS(status, "LALCreateUINT8FrequencySeries", FREQUENCYSERIESC);
	XLALPrintDeprecationWarning ("LALCreateUINT8FrequencySeries", "XLALCreateUINT8FrequencySeries");
	ASSERT(output != NULL, status, LAL_NULL_ERR, LAL_NULL_MSG);
	*output = XLALCreateUINT8FrequencySeries (name, &epoch, f0, deltaF, &sampleUnits, length);
	if(*output == NULL) {
		XLALClearErrno();
		ABORT(status, LAL_NOMEM_ERR, LAL_NOMEM_MSG);
	}
	RETURN(status);
}


UINT8FrequencySeries *XLALCutUINT8FrequencySeries (
	const UINT8FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALCutUINT8FrequencySeries";
	UINT8FrequencySeries *new;
	UINT8Sequence *sequence;

	new = XLALMalloc(sizeof(*new));
	sequence = XLALCutUINT8Sequence (series->data, first, length);
	if(!new || !sequence) {
		XLALFree(new);
		XLALDestroyUINT8Sequence (sequence);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}

	*new = *series;
	new->data = sequence;
	new->f0 += first * new->deltaF;

	return new;
}


UINT8FrequencySeries *XLALResizeUINT8FrequencySeries (
	UINT8FrequencySeries *series,
	int first,
	size_t length
)
{
	static const char func[] = "XLALResizeUINT8FrequencySeries";

	if(!XLALResizeUINT8Sequence (series->data, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	series->f0 += first * series->deltaF;

	return series;
}


UINT8FrequencySeries *XLALShrinkUINT8FrequencySeries (
	UINT8FrequencySeries *series,
	size_t first,
	size_t length
)
{
	static const char func[] = "XLALShrinkUINT8FrequencySeries";

	if(!XLALResizeUINT8FrequencySeries (series, first, length))
		XLAL_ERROR_NULL(func, XLAL_EFUNC);

	return series;
}


UINT8FrequencySeries *XLALAddUINT8FrequencySeries (
	UINT8FrequencySeries *arg1,
	const UINT8FrequencySeries *arg2
)
{
	static const char func[] = "XLALAddUINT8FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */
	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */
	if(Delta_f0 >= 0) {
		i = floor(Delta_f0 / arg1->deltaF + 0.5);
		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* add arg2 to arg1, adjusting the units */
	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		arg1->data->data[i] += arg2->data->data[j] * unit_ratio;
	}

	return arg1;
}





UINT8FrequencySeries *XLALMultiplyUINT8FrequencySeries (
	UINT8FrequencySeries *arg1,
	const UINT8FrequencySeries *arg2
)
{
	static const char func[] = "XLALMultiplyUINT8FrequencySeries";
	REAL8 Delta_f0 = arg2->f0 - arg1->f0;
	/* number of arg1 units per arg2 unit.  XLALUnitRatio() returns the
	 * number one obtains when one divides 1 of the first argument by 1
	 * of the second argument, for example if arg2 is in m and arg1 is
	 * in cm then unit_ratio = 100.0 */
	REAL8 unit_ratio = XLALUnitRatio(&arg2->sampleUnits, &arg1->sampleUnits);
	unsigned i, j;

	/* make sure arguments are compatible */

	if(XLALIsREAL8FailNaN(unit_ratio)) {
		XLALPrintError("%s(): incompatible sample units\n", func);
		XLAL_ERROR_NULL(func, XLAL_EFUNC);
	}
	if(XLALGPSCmp(&arg1->epoch, &arg2->epoch)) {
		XLALPrintError("%s(): incompatible epochs\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}
	if(arg1->deltaF != arg2->deltaF) {
		XLALPrintError("%s(): incompatible frequency resolution\n", func);
		XLAL_ERROR_NULL(func, XLAL_EDATA);
	}

	/* set start indexes */

	if(Delta_f0 >= 0) {
		unsigned start = floor(Delta_f0 / arg1->deltaF + 0.5);

		/* zero the first part of arg1 */

		for(i = 0; i < start && i < arg1->data->length; i++) {
			arg1->data->data[i] = 0.0;
		}

		j = 0;
	} else {
		i = 0;
		j = floor(-Delta_f0 / arg2->deltaF + 0.5);
	}

	/* multiply arg1 by arg2 */

	for(; i < arg1->data->length && j < arg2->data->length; i++, j++) {
		arg1->data->data[i] *= arg2->data->data[j] * unit_ratio;
	}

	/* zero the last part of arg1 */

	for(; i < arg1->data->length; i++) {
		arg1->data->data[i] = 0.0;
	}

	return arg1;
}

